
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "tk.h"

typedef struct {
	Tk_Window tkwin;
	Display *display;
	Tcl_Interp *interp;
	int x, y;
	int size;
	int borderWidth;
	Tk_3DBorder bgBorder;
	Tk_3DBorder fgBorder;
	int relief;
	GC gc;
	int updatePending;
} Square;

int SquareCmd(ClientData clientData, Tcl_Interp *interp,
	int argc, char *argv[]);
 
int SquareConfigure(Tcl_Interp *interp, Square *squarePtr,
	int argc, char *argv[], int flags);

int SquareWidgetCmd(ClientData clientData, Tcl_Interp *interp,
	int argc, char *argv[]);

void SquareDestroy(ClientData clientData);

void SquareDisplay(ClientData clientData);

void SquareEventProc(ClientData clientData, XEvent *eventPtr);

void KeepInWindow(Square *squarePtr);

int SquareWidgetCmd(ClientData clientData, Tcl_Interp *interp,
	int argc, char *argv[]);

int Square_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp,"square",SquareCmd,Tk_MainWindow(interp),NULL);
}



int SquareCmd(ClientData clientData, Tcl_Interp *interp,
		int argc, char *argv[]) {
	Tk_Window main = (Tk_Window) clientData;
	Square *squarePtr;
	Tk_Window tkwin;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				argv[0], " pathName ?options?\"", (char *) NULL);
		return TCL_ERROR;
	}

	tkwin = Tk_CreateWindowFromPath(interp, main, argv[1],
			(char *) NULL);
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	Tk_SetClass(tkwin, "Square");

	squarePtr = (Square *) malloc(sizeof(Square));
	squarePtr->tkwin = tkwin;
	squarePtr->display = Tk_Display(tkwin);
	squarePtr->interp = interp;
	squarePtr->x = squarePtr->y = 0;
	squarePtr->size = 30;
	squarePtr->borderWidth = 0;
	squarePtr->bgBorder = squarePtr->fgBorder = NULL;
	squarePtr->relief = TK_RELIEF_FLAT;
     // dlaczego w tym miejscu nadaje sie wartosci skladowym struktury ?
	squarePtr->gc = None;
	squarePtr->updatePending = 0;

	Tk_CreateEventHandler(tkwin,
		ExposureMask|StructureNotifyMask, SquareEventProc,
		(ClientData) squarePtr);
	Tcl_CreateCommand(interp, Tk_PathName(tkwin), 
		SquareWidgetCmd, (ClientData) squarePtr,
		(Tcl_CmdDeleteProc *) NULL);
	if (SquareConfigure(interp, squarePtr, argc-2, argv+2, 0)
			!= TCL_OK) {
		Tk_DestroyWindow(squarePtr->tkwin);
		return TCL_ERROR;
	}
	interp->result = Tk_PathName(tkwin);
	return TCL_OK;
}


Tk_ConfigSpec configSpecs[] = {
	{TK_CONFIG_BORDER, "-background", "background",
		"Background", "#cdb79e",
		Tk_Offset(Square, bgBorder), TK_CONFIG_COLOR_ONLY,
		(Tk_CustomOption *) NULL},
	{TK_CONFIG_BORDER, "-background", "background",
		"Background", "white",
      Tk_Offset(Square, bgBorder), TK_CONFIG_MONO_ONLY,
      (Tk_CustomOption *) NULL},
	{TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
		(char *) NULL, 0, 0, (Tk_CustomOption *) NULL},
	{TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
		(char *) NULL, 0, 0, (Tk_CustomOption *) NULL},
	{TK_CONFIG_PIXELS, "-borderwidth", "borderWidth",
		"BorderWidth", "0.3c", Tk_Offset(Square, borderWidth),
		0, (Tk_CustomOption *) NULL},
	{TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
		(char *) NULL, 0, 0, (Tk_CustomOption *) NULL},
	{TK_CONFIG_BORDER, "-foreground", "foreground",
		"Foreground", "#b03060",
		Tk_Offset(Square, fgBorder), TK_CONFIG_COLOR_ONLY,
		(Tk_CustomOption *) NULL},
	{TK_CONFIG_BORDER, "-foreground", "foreground",
		"Foreground", "black",
		Tk_Offset(Square, fgBorder), TK_CONFIG_MONO_ONLY,
		(Tk_CustomOption *) NULL},
	{TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
		"groove", Tk_Offset(Square, relief), 0,
		(Tk_CustomOption *) NULL},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL,
		(char *) NULL, (char *) NULL, 0, 0,
		(Tk_CustomOption *) NULL}
};


int SquareConfigure(Tcl_Interp *interp, Square *squarePtr,
		int argc, char *argv[], int flags) {
	if (Tk_ConfigureWidget(interp, squarePtr->tkwin, configSpecs,
			argc, argv, (char *) squarePtr, flags) != TCL_OK) {
		return TCL_ERROR;
	}
	Tk_SetWindowBackground(squarePtr->tkwin,
			Tk_3DBorderColor(squarePtr->bgBorder)->pixel);
	if (squarePtr->gc == None) {
		XGCValues gcValues;
		gcValues.function = GXcopy;
		gcValues.graphics_exposures = False;
		squarePtr->gc = Tk_GetGC(squarePtr->tkwin,
				GCFunction|GCGraphicsExposures, &gcValues);
	}
	Tk_GeometryRequest(squarePtr->tkwin, 200, 150);
	Tk_SetInternalBorder(squarePtr->tkwin,
			squarePtr->borderWidth);
	if (!squarePtr->updatePending) {
		Tk_DoWhenIdle(SquareDisplay, (ClientData) squarePtr);
		squarePtr->updatePending = 1;
	}
	return TCL_OK;
}


int SquareWidgetCmd(ClientData clientData, Tcl_Interp *interp,
		int argc, char *argv[]) {
	Square *squarePtr = (Square *) clientData;
	int result = TCL_OK;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				argv[0], " option ?arg arg ...?\"",
				(char *) NULL);
		return TCL_ERROR;
	}

	Tk_Preserve((ClientData) squarePtr);
	if (strcmp(argv[1], "configure") == 0) {
		if (argc == 2) {
			result = Tk_ConfigureInfo(interp, squarePtr->tkwin,
					configSpecs, (char *) squarePtr, (char *) NULL, 0);
		} else if (argc == 3) {
			result = Tk_ConfigureInfo(interp, squarePtr->tkwin,
					configSpecs, (char *) squarePtr, argv[2], 0);
		} else {
			result = SquareConfigure(interp, squarePtr, argc-2,
					argv+2, TK_CONFIG_ARGV_ONLY);
		}
	} else if (strcmp(argv[1], "position") == 0) {
		if ((argc != 2) && (argc != 4)) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					argv[0], " position ?x y?\"", (char *) NULL);
			goto error;
		}
		if (argc == 4) {
			if ((Tk_GetPixels(interp, squarePtr->tkwin, argv[2],
					&squarePtr->x) != TCL_OK) ||
					(Tk_GetPixels(interp, squarePtr->tkwin,
					argv[3], &squarePtr->y) != TCL_OK)) {
				goto error;
			}
			KeepInWindow(squarePtr);
		}
		sprintf(interp->result, "%d %d", squarePtr->x, squarePtr->y);
      /*
          dlaczego mozna pisac bezposrednio do "interp->result" ???
          gdyz standardowo interp-result wskazuje na pusty string
          znajdujacy sie w tablicy TCL_RESULT_SIZE-elementowej !!!
          patrz: help Tcl_Interp
      */
         

	} else if (strcmp(argv[1], "size") == 0) {
	if ((argc != 2) && (argc != 3)) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					argv[0], " size ?amount?\"", (char *) NULL);
			goto error;
		}
		if (argc == 3) {
			int i;
			if (Tk_GetPixels(interp, squarePtr->tkwin, argv[2],
					&i) != TCL_OK) {
				goto error;
			}
			if ((i <= 0) || (i > 100)) {
				Tcl_AppendResult(interp, "bad size \"", argv[2],
						"\"", (char *) NULL);
				goto error;
			}
			squarePtr->size = i;
			KeepInWindow(squarePtr);
		}
		sprintf(interp->result, "%d", squarePtr->size);
	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
				"\": must be configure, position, or size",
				(char *) NULL);
		goto error;
	}
	if (!squarePtr->updatePending) {
		Tk_DoWhenIdle(SquareDisplay, (ClientData) squarePtr);
		squarePtr->updatePending = 1;
	}
	Tk_Release((ClientData) squarePtr);
	return result;

error:
	Tk_Release((ClientData) squarePtr);
	return TCL_ERROR;
}



void KeepInWindow(Square *squarePtr) {
	int gap, bd;
	bd = 0;
	if (squarePtr->relief != TK_RELIEF_FLAT) {
		bd = squarePtr->borderWidth;
	}
	gap = (Tk_Width(squarePtr->tkwin) - bd)
			- (squarePtr->x + squarePtr->size);
	if (gap < 0) {
		squarePtr->x += gap;
	}
	gap = (Tk_Height(squarePtr->tkwin) - bd)
			- (squarePtr->y + squarePtr->size);
	if (gap < 0) {
		squarePtr->y += gap;
	}
	if (squarePtr->x < bd) {
		squarePtr->x = bd;
	}
	if (squarePtr->y < bd) {
		squarePtr->y = bd;
	}
}


void SquareEventProc(ClientData clientData, XEvent *eventPtr) {
	Square *squarePtr = (Square *) clientData;
	if (eventPtr->type == Expose) {
		if (!squarePtr->updatePending) {
			Tk_DoWhenIdle(SquareDisplay, (ClientData) squarePtr);
			squarePtr->updatePending = 1;
		}
	} else if (eventPtr->type == ConfigureNotify) {
		KeepInWindow(squarePtr);
		if (!squarePtr->updatePending) {
			Tk_DoWhenIdle(SquareDisplay, (ClientData) squarePtr);
			squarePtr->updatePending = 1;
		}
	} else if (eventPtr->type == DestroyNotify) {
		Tcl_DeleteCommand(squarePtr->interp,
				Tk_PathName(squarePtr->tkwin));
		squarePtr->tkwin = NULL;
		if (squarePtr->updatePending) {
			Tk_CancelIdleCall(SquareDisplay,
					(ClientData) squarePtr);
		}
		Tk_EventuallyFree((ClientData) squarePtr,
         (Tcl_FreeProc*)SquareDestroy);
        // dodalem (Tcl_FreeProc*)
	}
}


void SquareDisplay(ClientData clientData) {
	Square *squarePtr = (Square *) clientData;
	Tk_Window tkwin = squarePtr->tkwin;
	Display *display = Tk_Display(tkwin);
	Pixmap pm;
	squarePtr->updatePending = 0;
	if (!Tk_IsMapped(tkwin)) {
		return;
	}

   //
   // to musimy zmienic w Tcl/Tk 8.0 !!!
   //

	//pm = XCreatePixmap(display, Tk_WindowId(tkwin),
	//		Tk_Width(tkwin), Tk_Height(tkwin), Tk_Depth(tkwin));
	pm = Tk_GetPixmap(display, Tk_WindowId(tkwin),
			Tk_Width(tkwin), Tk_Height(tkwin), Tk_Depth(tkwin));
   /*
	Tk_Fill3DRectangle(display, pm, squarePtr->bgBorder,
		0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
		squarePtr->borderWidth, squarePtr->relief);
	Tk_Fill3DRectangle(display, pm, squarePtr->fgBorder,
		squarePtr->x, squarePtr->y, squarePtr->size,
		squarePtr->size, squarePtr->borderWidth,
		squarePtr->relief);
   */
	Tk_Fill3DRectangle(tkwin, pm, squarePtr->bgBorder,
		0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
		squarePtr->borderWidth, squarePtr->relief);
	Tk_Fill3DRectangle(tkwin, pm, squarePtr->fgBorder,
		squarePtr->x, squarePtr->y, squarePtr->size,
		squarePtr->size, squarePtr->borderWidth,
		squarePtr->relief);
	XCopyArea(display, pm, Tk_WindowId(tkwin),
		squarePtr->gc, 0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
		0, 0);
	//XFreePixmap(Tk_Display(tkwin), pm);
   Tk_FreePixmap(Tk_Display(tkwin), pm);
}


void SquareDestroy(ClientData clientData) {
	Square *squarePtr = (Square *) clientData;
	Tk_FreeOptions(configSpecs, (char *) squarePtr,
			squarePtr->display, 0);
	if (squarePtr->gc != None) {
		Tk_FreeGC(squarePtr->display, squarePtr->gc);
	}
	free((char *) squarePtr);
}

